IMPLEMENTATION MODULE Server;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Str, Lib, Lib2, SO, Crc, MD5, Tmrs, Tmr, UFTP, Config, CliIO, Files, Sndrs, Rcvrs, Utils, Cfg, Log;

CONST
  mid = Cfg.midServer;

  dtRnd = 20;            (* ms, estimated roundtrip time *)
  toDat = 10 * dtRnd;    (* ms *)
  toSes = 30 * 1000;     (* ms *)

  fglMax = UFTP.lenMaxFgm;  (* adjust here *)
  dtuMax = 1000;            (* ms *)
  fpuMax = 100;

  cntMax = Cfg.cnt;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. cntMax ];

  TSndDesc = RECORD
    f: Files.TFile;
    s: Sndrs.TSndr;
  END;

  TRcvDesc = RECORD
    f: Files.TFile;
    r: Rcvrs.TRcvr;
  END;

  TExtDesc = RECORD
    CASE w (* wrt *): BOOLEAN OF
    | FALSE: s: TSndDesc;
    | TRUE : r: TRcvDesc;
    END;
  END;

  TParDesc = RECORD
    t: CARDINAL;     (* dts (session timeout) *)
    c, m: CARDINAL;  (* dtu cnt, max *)
    n: CARDINAL;     (* fpu *)
  END;

  TSesDesc = RECORD
    i: TIdx;            (* 0 <==> free *)
    d: CliIO.TDstDesc;  (* = session ID *)
    t: CARDINAL;
    x: TExtDesc;
    p: TParDesc;
  END;
  TSessDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TSesDesc;
  END;

VAR
  ses: TSessDesc;
  ers: TErrCnts;
  tmr: Tmrs.TTimer;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* solve return code problem before removing this line *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE IncErrCnt( e: TErr );
BEGIN
  Lib2.IncL( ers[e], 1 );
  Lib2.IncL( ers[eNull],1  );
END IncErrCnt;


PROCEDURE Alloc( d-: CliIO.TDstDesc; x-: TExtDesc; p-: TParDesc ): TIdx;
VAR
  i: TIdx;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
    r: UFTP.TResID;
  BEGIN
    IF ~ ena THEN RETURN END;
    IF x.w THEN
      Files.GetRID( x.r.f, r );
    ELSE
      Files.GetRID( x.s.f, r );
    END;
    SO.Init( b, s );
    SO.WrStr( b, 'Alloc idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' dst ' );
    Utils.WrIP( b, d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, d.p, 1 );
    SO.WrStr( b, ' wrt ' );
    SO.WrCard( b, ORD( x.w ), 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, r );            (* not really a string, check !!! *)
    SO.WrStr( b, "' dts " );
    SO.WrLngCard( b, p.t, 1 );
    SO.WrStr( b, ' dtu ' );
    SO.WrLngCard( b, p.m, 1 );
    SO.WrStr( b, ' fpu ' );
    SO.WrLngCard( b, p.n, 1 );
    Message( s );
  END Show;

BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    IF ses.a[i].i = 0 THEN
      ses.a[i].i := i;
      ses.a[i].d := d;
      ses.a[i].x := x;
      ses.a[i].p := p;
      INC( ses.n );
      Show();
      RETURN i;
    END;
  END;
  RETURN 0;
END Alloc;

PROCEDURE Free( i: TIdx );

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
    r: UFTP.TResID;
  BEGIN
    IF ~ ena THEN RETURN END;
    IF ses.a[i].x.w THEN
      (*
      Rcvrs.GetRID( ses.a[i].x.r.f, r );
      *)
      Lib2.Fill( r, 0 );
    ELSE
      Files.GetRID( ses.a[i].x.s.f, r );
    END;
    SO.Init( b, s );
    SO.WrStr( b, 'Free idx ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' dst ' );
    Utils.WrIP( b, ses.a[i].d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, ses.a[i].d.p, 1 );
    SO.WrStr( b, ' wrt ' );
    SO.WrCard( b, ORD( ses.a[i].x.w ), 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, r );
    SO.WrStr( b, "'" );
    Message( s );
  END Show;

VAR
  t, p: UFTP.TPosDesc;
BEGIN
  IF i = 0 THEN RETURN END;
  IF ses.a[i].i # i THEN Fatal( 'Free inconsistency' ) END;
  Show();
  IF ses.a[i].x.w THEN
    Rcvrs.Free( ses.a[i].x.r.r );
    Files.GetTgt( ses.a[i].x.r.f, t );
    Files.GetPos( ses.a[i].x.r.f, 0, p );
    IF Lib2.Compare( t, p ) # SIZE( p ) THEN Files.Trunc( ses.a[i].x.r.f, 0 ) END;
    Files.Close( ses.a[i].x.r.f );
  ELSE
    Sndrs.Free( ses.a[i].x.s.s );
    Files.Close( ses.a[i].x.s.f );
  END;
  DEC( ses.n );
  Lib2.Fill( ses.a[i], 0 )
END Free;

PROCEDURE Find( d-: CliIO.TDstDesc ): TIdx;
VAR
  i: TIdx;
BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    IF ( ses.a[i].i > 0 ) & ( Lib2.Compare( ses.a[i].d, d ) = SIZE( d ) ) THEN RETURN i END;
  END;
  RETURN 0;
END Find;


PROCEDURE OnRx( rxi: CliIO.TRxInfDesc; dat-: ARRAY OF BYTE; len: CARDINAL );
VAR
  tel: POINTER TO UFTP.TTelDesc;

  PROCEDURE RdRq();
  VAR
    i: TIdx;
    x: TExtDesc;
    p: TParDesc;

    PROCEDURE Reply( e: UFTP.TError );
    VAR
      rep: RECORD h: UFTP.THdrDesc; b: UFTP.TRdRpDesc; t: UFTP.TTrlDesc END;
    BEGIN
      Lib2.Fill( rep, 0 );
      rep.h.m := UFTP.mag;
      rep.h.v := UFTP.ver;
      rep.h.c := UFTP.cmdRdRp;
      rep.b.e := e;
      IF e = UFTP.errNull THEN
        Sndrs.GetFgl( x.s.s, rep.b.f );
        Files.GetTgt( x.s.f, rep.b.p );
        rep.b.r := UFTP.TReference( i );
      END;
      rep.t := Crc.Swap( Crc.UpdL( Crc.crcInit, rep, SIZE(rep) - SIZE(rep.t) ) );
      CliIO.Send( rxi.d, rep, SIZE( rep ) );
    END Reply;

  CONST
    lenMin = SIZE( UFTP.TRdRqDesc );
    lenMax = lenMin;
  VAR
    l, m: UFTP.TFgmLen;
    r: UFTP.TResID;
  BEGIN
    i := 0;
    Lib2.Fill( x, 0 );
    Lib2.Fill( p, 0 );
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eRdRqLen );
      RETURN;
    END;
    l := tel^.rrq.f;
    IF l = 0 THEN
      IncErrCnt( eRdRqFgl );
      RETURN;
    END;
    IF l > fglMax THEN l := fglMax END;
    i := Find( rxi.d );
    IF i > 0 THEN
      x := ses.a[i].x;
      p := ses.a[i].p;
      IF x.w THEN
        IncErrCnt( eRdRqTyp );
        RETURN;
      END;
      Files.GetRID( x.s.f, r );
      Sndrs.GetFgl( x.s.s, m );
      IF ( Str.Compare( tel^.rrq.r, r ) = 0 ) & ( l = m ) THEN
        Reply( UFTP.errNull );
      ELSE
        IncErrCnt( eRdRqPar );
      END;
      RETURN;
    END;
    IF ses.n >= cntMax THEN
      IncErrCnt( eRdRqRes );
      Reply( UFTP.errRes );
      RETURN;
    END;
    x.w := FALSE;
    x.s.f := Files.OpenRd( tel^.rrq.r );
    IF x.s.f = Files.fNull THEN
      IncErrCnt( eRdRqRID );
      Reply( UFTP.errRID );
      RETURN;
    END;
    x.s.s := Sndrs.Alloc( l, x.s.f );
    IF x.s.s = Sndrs.sNull THEN Fatal( 'RdRq snd' ) END;
    p.t := toSes;
    p.m := Config.data.dtu;
    p.n := Config.data.fpu;
    i := Alloc( rxi.d, x, p );
    IF i = 0 THEN Fatal( 'RdRq ses' ) END;
    Reply( UFTP.errNull );
  END RdRq;

  PROCEDURE DatRq();
  CONST
    lenMin = SIZE( UFTP.TDatRqDesc );
    lenMax = lenMin;
  VAR
    r: UFTP.TReference;
    i: TVldIdx;
  BEGIN
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRqLen );
      RETURN;
    END;
    r := tel^.drq.h.r;
    IF ( r < MIN( TVldIdx ) ) OR ( r > MAX( TVldIdx ) ) THEN
      IncErrCnt( eDatRqRef );
      RETURN;
    END;
    i := TVldIdx( r );
    IF ( ses.a[i].i # i ) OR ( Lib2.Compare( ses.a[i].d, rxi.d ) < SIZE( rxi.d ) ) OR ses.a[i].x.w THEN
      IncErrCnt( eDatRqSes );
      RETURN;
    END;
    ses.a[i].t := toSes;
    Sndrs.Receive( ses.a[i].x.s.s, tel^.drq.h.i, tel^.drq.s );
  END DatRq;

  PROCEDURE Close();
  CONST
    lenMin = SIZE( UFTP.TCloseDesc );
    lenMax = lenMin;
  VAR
    r: UFTP.TReference;
    i: TVldIdx;
  BEGIN
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eCloseLen );
      RETURN;
    END;
    r := tel^.cls.r;
    IF ( r < MIN( TVldIdx ) ) OR ( r > MAX( TVldIdx ) ) THEN
      IncErrCnt( eCloseRef );
      RETURN;
    END;
    i := TVldIdx( r );
    IF ( ses.a[i].i # i ) OR ( Lib2.Compare( ses.a[i].d, rxi.d ) < SIZE( rxi.d ) ) THEN
      IncErrCnt( eCloseSes );
      RETURN;
    END;
    Free( i );
  END Close;

  PROCEDURE WrRq();
  CONST
    lenMin = SIZE( UFTP.TWrRqDesc );
    lenMax = lenMin;
  VAR
    i: TIdx;
    l: UFTP.TFgmLen;
    p: TParDesc;

    PROCEDURE Reply( e: UFTP.TError );
    VAR
      rep: RECORD h: UFTP.THdrDesc; b: UFTP.TWrRpDesc; t: UFTP.TTrlDesc END;
    BEGIN
      Lib2.Fill( rep, 0 );
      rep.h.m := UFTP.mag;
      rep.h.v := UFTP.ver;
      rep.h.c := UFTP.cmdWrRp;
      rep.b.e := e;
      IF e = UFTP.errNull THEN
        rep.b.f := l;
        rep.b.r := UFTP.TReference( i );
      END;
      rep.t := Crc.Swap( Crc.UpdL( Crc.crcInit, rep, SIZE(rep) - SIZE(rep.t) ) );
      CliIO.Send( rxi.d, rep, SIZE( rep ) );
    END Reply;

  VAR
    x: TExtDesc;
    r: UFTP.TResID;
    t: UFTP.TPosDesc;
    m: UFTP.TFgmLen;
  BEGIN
    i := 0;
    Lib2.Fill( p, 0 );
    Lib2.Fill( x, 0 );
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eWrRqLen );
      RETURN;
    END;
    l := tel^.wrq.f;
    IF l = 0 THEN
      IncErrCnt( eWrRqFgl );
      RETURN;
    END;
    IF l > fglMax THEN l := fglMax END;
    i := Find( rxi.d );
    IF i > 0 THEN
      x := ses.a[i].x;
      IF ~ x.w THEN
        IncErrCnt( eWrRqTyp );
        RETURN;
      END;
      Files.GetRID( x.r.f, r );
      Files.GetTgt( x.r.f, t );
      Rcvrs.GetFgl( x.r.r, m );
      IF ( Str.Compare( tel^.wrq.r, r ) = 0 ) & ( Lib2.Compare( tel^.wrq.p, t ) = SIZE( t ) ) & ( l = m ) THEN
        Reply( UFTP.errNull );
      ELSE
        IncErrCnt( eWrRqPar );
      END;
      RETURN;
    END;
    IF ses.n >= cntMax THEN
      IncErrCnt( eWrRqRes );
      Reply( UFTP.errRes );
      RETURN;
    END;
    x.w := TRUE;
    x.r.f := Files.OpenWr( tel^.wrq.r );
    IF x.r.f = Files.fNull THEN
      IncErrCnt( eWrRqRID );
      Reply( UFTP.errRID );
      RETURN;
    END;
    Files.SetTgt( x.r.f, tel^.wrq.p );
    p.t := toSes;
    p.m := Config.data.dtu;  (* !!! use limited params from req *)
    p.n := Config.data.fpu;
    x.r.r := Rcvrs.Alloc( l, x.r.f, p.n * dtRnd DIV p.m );
    IF x.r.r = Rcvrs.rNull THEN Fatal( 'WrRq rcv' ) END;
    i := Alloc( rxi.d, x, p );
    IF i = 0 THEN Fatal( 'WrRq ses' ) END;
    Reply( UFTP.errNull );
  END WrRq;

  PROCEDURE DatRp();
  CONST
    lenMin = UFTP.lenMinDatRpDesc;
    lenMax = UFTP.lenMaxDatRpDesc;
  VAR
    r: UFTP.TReference;
    i: TVldIdx;
  BEGIN
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRpLen );
      RETURN;
    END;
    DEC( len, SIZE( UFTP.TDatHdrDesc ) );
    r := tel^.drp.h.r;
    IF ( r < MIN( TVldIdx ) ) OR ( r > MAX( TVldIdx ) ) THEN
      IncErrCnt( eDatRpRef );
      RETURN;
    END;
    i := TVldIdx( r );
    IF ( ses.a[i].i # i ) OR ( Lib2.Compare( ses.a[i].d, rxi.d ) < SIZE( rxi.d ) ) OR ~ ses.a[i].x.w THEN
      IncErrCnt( eDatRpSes );
      RETURN;
    END;
    ses.a[i].t := toSes;
    Rcvrs.Receive( ses.a[i].x.r.r, tel^.drp.h.i, tel^.drp.d, len );
  END DatRp;

CONST
  lenMin = UFTP.lenFrmOvr;
  lenMax = UFTP.lenMaxTel;
BEGIN
  IF ( len < lenMin ) OR ( len > lenMax ) THEN IncErrCnt( eLen ); RETURN END;
  IF ~ Crc.Ok( dat, len ) THEN IncErrCnt( eCrc ); RETURN END;
  tel := ADR( dat );
  DEC( len, lenMin );
  IF tel^.hdr.m # UFTP.mag THEN IncErrCnt( eMag ); RETURN END;
  IF tel^.hdr.v # UFTP.ver THEN IncErrCnt( eVer ); RETURN END;
  CASE tel^.hdr.c OF
  | UFTP.cmdRdRq : RdRq();
  | UFTP.cmdDatRq: DatRq();
  | UFTP.cmdClose: Close();
  | UFTP.cmdWrRq : WrRq();
  | UFTP.cmdDatRp: DatRp();
  ELSE
    IncErrCnt( eCmd );
  END;
END OnRx;


PROCEDURE DatRq( r: Rcvrs.TRef; i: UFTP.TFgmIdx; s: UFTP.TFgmSet );
VAR
  t: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRqDesc; t: UFTP.TTrlDesc END;
BEGIN
  Lib2.Fill( t, 0 );
  t.h.m := UFTP.mag;
  t.h.v := UFTP.ver;
  t.h.c := UFTP.cmdDatRq;
  t.b.h.r := UFTP.TReference( r );
  t.b.h.i := i;
  t.b.s := s;
  t.t := Crc.Swap( Crc.UpdL( Crc.crcInit, t, SIZE(t) - SIZE(t.t) ) );
  CliIO.Send( ses.a[TVldIdx(r)].d, t, SIZE(t) );
END DatRq;

PROCEDURE AppendCrc( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
TYPE
  TCrcPtr = POINTER TO Crc.TCrc;
BEGIN
  TCrcPtr( ADR(x[l]) )^ := Crc.Swap( Crc.UpdL( Crc.crcInit, x, l ) );
  RETURN l + SIZE( Crc.TCrc );
END AppendCrc;

PROCEDURE DatRp( r: Sndrs.TRef; i: UFTP.TFgmIdx; x-: ARRAY OF BYTE; l: CARDINAL );
VAR
  t: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRpDesc; t: UFTP.TTrlDesc END;
BEGIN
  Lib2.Fill( t, 0 );
  t.h.m := UFTP.mag;
  t.h.v := UFTP.ver;
  t.h.c := UFTP.cmdDatRp;
  t.b.h.r := 0;
  t.b.h.i := i;
  Lib.Move( ADR(x), ADR(t.b.d), l );
  l := AppendCrc( t, SIZE(t.h) + SIZE(t.b.h) + l );
  CliIO.Send( ses.a[TVldIdx(r)].d, t, l );
END DatRp;

PROCEDURE Close( i: TVldIdx );
VAR
  t: RECORD h: UFTP.THdrDesc; b: UFTP.TCloseDesc; t: UFTP.TTrlDesc END;
BEGIN
  Lib2.Fill( t, 0 );
  t.h.m := UFTP.mag;
  t.h.v := UFTP.ver;
  t.h.c := UFTP.cmdClose;
  t.b.r := 0;
  t.t := Crc.Swap( Crc.UpdL( Crc.crcInit, t, SIZE(t) - SIZE(t.t) ) );
  CliIO.Send( ses.a[i].d, t, SIZE( t ) );

  Free( i );
END Close;

PROCEDURE OnTmr( k: Tmrs.TTicks; r: Tmrs.TRef );
VAR
  i: TVldIdx;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'OnTmr idx ' );
    SO.WrCard( b, i, 1 );
    Message( s );
  END Show;

VAR
  s: POINTER TO TSesDesc;
BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    s := ADR( ses.a[i] );
    IF s^.i > 0 THEN
      IF s^.p.t > 0 THEN
        DEC( s^.p.t );
        IF s^.p.c > 0 THEN DEC( s^.p.c ) END;
        IF s^.p.c = 0 THEN
          s^.p.c := s^.p.m;
          IF s^.x.w THEN
            IF ~ Rcvrs.Update( s^.x.r.r, s^.p.n, toDat DIV s^.p.m, DatRq, Rcvrs.TRef(i) ) THEN Close( i ) END;
          ELSE
            Sndrs.Update( s^.x.s.s, s^.p.n, DatRp, Sndrs.TRef(i) );
          END;
        END;
      ELSE
        Show();
        Free( i );
      END;
    END;
  END;
END OnTmr;


PROCEDURE Run();
BEGIN
  CliIO.Execute();
END Run;

PROCEDURE GetErrCnts( VAR e: TErrCnts );
BEGIN
  e := ers;
END GetErrCnts;


PROCEDURE InitModule();

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..128] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'InitModule UFTP ver ' );
    SO.WrCard( b, UFTP.ver, 1 );
    SO.WrStr( b, ' lenMaxTel ' );
    SO.WrCard( b, UFTP.lenMaxTel, 1 );
    SO.WrStr( b, ' lenFrmOvr ' );
    SO.WrCard( b, UFTP.lenFrmOvr, 1 );
    SO.WrStr( b, ' lenMaxVar ' );
    SO.WrCard( b, UFTP.lenMaxVar, 1 );
    SO.WrStr( b, ' lenMaxFgm ' );
    SO.WrCard( b, UFTP.lenMaxFgm, 1 );
    Message( s );
    SO.Init( b, s );
    SO.WrStr( b, 'InitModule cnt ' );
    SO.WrCard( b, Cfg.cnt, 1 );
    Message( s );
  END Show;

BEGIN
  Log.Register( mid, 'Server' );
  Lib2.Fill( ses, 0 );
  Lib2.Fill( ers, 0 );

  tmr := Tmrs.New();
  Tmrs.Set( tmr, 1, 1 );  (* fix 1ms *)
  Tmrs.SetHnd( tmr, OnTmr, 0 );

  CliIO.Install( OnRx );

  Show();
END InitModule;

BEGIN
  InitModule();
END Server.
